@@ -257,8 +257,10 @@ class Matches:
 
     Raises
     ------
-    ValueError: Filename did not match pattern.
-    IndexError: Not as many matches as matchers.
+    ValueError
+        Filename did not match pattern.
+    IndexError
+        Not as many matches as matchers.
     """
 
     def __init__(self, matchers: List[Matcher], filename: str,
@@ -296,6 +298,7 @@ class Matches:
         See :func:`Finder.get_matchers
         <filefinder.finder.Finder.get_matchers>` for details on
         `key` argument.
+        :func:`__getitem__` wraps around this method.
 
         Returns
         -------
@@ -318,8 +321,10 @@ def get_matchers_indices(matchers: List[Matcher],
 
     Raises
     ------
-    IndexError: No matcher found corresponding to the key
-    TypeError: Key is not int or str
+    IndexError
+        No matcher found corresponding to the key
+    TypeError
+        Key is not int or str
     """
     if isinstance(key, int):
         return [key]
